open Language
open Environment
open Display

exception Kernel_panic of string

module type Kernel =
sig
  val run : obj -> obj
end

module Kernel (Env : Environment) : Kernel =
struct
  open Env

  type continuation = environment -> obj -> unit

  let abs_continuation_of_continuation (cont : continuation) : abs_continuation = Obj.magic cont
  let continuation_of_abs_continuation (cont : abs_continuation) : continuation = Obj.magic cont

  let err obj str = raise (Kernel_panic ("Error in " ^ (quote_string (string_of_obj obj)) ^ " : " ^ str ^ "\n"))

  let int_of_obj = function
    | Entier i -> i
    | o -> err o "couldn't evaluate object as an integer"

  let rec exec code env cont =
    let err = err code in
      match code with
	| Entier _ as e -> cont env e
	| Chaine c ->
	    (*
	      try*)
		cont (get_environment env c) (get_value env c)
	      (*with
		| Not_found -> err "trying to get the contents of an unbound variable"
	    *)
	| Liste([]) as e -> cont env e
	| Liste(Op_add::[]) -> cont env (Entier 0)
	| Liste(Op_add::(Entier i)::s) -> exec (Liste(Op_add::s)) env (fun env r -> cont env (Entier(i+(int_of_obj r))))
	| Liste(Op_add::a::s) -> exec a env (fun _ -> function 
					       | Entier i -> exec (Liste(Op_add::Entier i::s)) env cont
					       | r -> err ("trying to add " ^ (quote_string_of_obj r) ^ " which is not an integer")
					    )
	| Liste(Op_sub::[a]) ->
	    exec a env (fun _ -> function
			  | Entier i -> cont env (Entier (-i))
			  | r -> err ("trying to add " ^ (quote_string_of_obj r) ^ " which is not an integer")
		       )
	| Liste(Op_sub::a::s) -> exec (Liste(Op_add::a::[Liste(Op_sub::[Liste(Op_add::s)])])) env cont
	| Liste(Op_div::a::[b]) -> exec a env (fun _ -> function
						 | Entier p -> exec b env
						     (fun _ -> function
							| Entier 0 -> err "trying to divide by zero"
							| Entier q -> cont env (Entier (p/q))
							| _ -> err "you can only divide integers"
						     )
						 | _ -> err "you can only divide integers")
	| Liste(Op_div::_) -> err "wrong number of arguments"
	| Liste(Op_mult::[]) -> cont env (Entier 1)
	| Liste(Op_mult::(Entier i)::s) -> exec (Liste(Op_mult::s)) env (fun _ r -> cont env (Entier(i*(int_of_obj r))))
	| Liste(Op_mult::a::s) -> exec a env (fun _ -> function
						| Entier i as r -> exec (Liste(Op_mult::r::s)) env cont
						| r -> err ("trying to multiply " ^ (quote_string_of_obj r) ^ " which is not an integer")
					     )
	| Liste(Op_quote::[q]) -> cont env q
	| Liste(Op_quote::_) -> err "wrong number of arguments"
	| Liste(Op_cons::a::[b]) ->
	    exec a env (fun _ ra -> exec b env (fun _ rb -> cont env (Liste(ra::[rb]))))
	| Liste(Op_cons::_) -> err "wrong number of arguments"
	| Liste(Op_let::Liste(Liste(Chaine c::[v])::s)::[op]) ->
	    exec v env (fun _ r -> exec (Liste(Op_let::Liste(s)::[op])) (link env c r) cont)
	| Liste(Op_let::Liste([])::[op]) -> exec op env cont
	| Liste(Op_let::_) -> err "wrong number or type of arguments"
	| Liste(Op_set::x::[v]) ->
	    exec x env (fun _ r -> exec (Liste(Op_setq::r::[v])) env cont)
	| Liste(Op_set::_) -> err "wrong number of arguments"
	| Liste(Op_setq::Chaine c::[v]) ->
	    exec v env (fun _ r -> ignore (relink env c r); cont env (get_value env c))
	| Liste(Op_setq::_) -> err "wrong number of arguments"
	| Liste(Op_eval::[o]) -> exec o env (fun env r -> exec r env cont)
	| Liste(Op_eval::_) -> err "wrong number of arguments"
	| Liste(Op_print::[o]) -> exec o env (fun _ r -> Display.print_obj r; Printf.printf "\n"; cont env r)
	| Liste(Op_print::_) -> err "wrong number of arguments"
	| Liste(Op_callcc::[f]) ->
	    exec (Liste(f::[Continuation (abs_environment_of_environment env, abs_continuation_of_continuation cont)])) env cont
	| Liste(Op_callcc::_) -> err "wrong number of arguments"
	| Liste(Continuation(e, cont)::[o]) ->
	    exec o env (fun _ r -> (continuation_of_abs_continuation cont) (environment_of_abs_environment e) r) 
	| Liste(Continuation(_, _)::_) -> err "wrong number of arguments"
	| Continuation(e, c) -> cont env (Continuation (e, c))
	| Liste(Op_begin::[a]) -> exec a env cont
	| Liste(Op_begin::a::s) ->
	    exec a env (fun _ _ -> exec (Liste(Op_begin::s)) env cont)
	| Liste(Op_is_cons::[o]) ->
	    exec o env (fun _ r ->
			  match r with
			    | Liste(_::_) -> cont env (Liste([Entier 1]))
			    | _ -> cont env (Liste([]))
		       )
	| Liste(Op_is_cons::_) -> err "wrong number of arguments"
	| Liste(Op_car::[l]) -> exec l env (fun _ -> function
					      | Liste(l) -> cont env (List.hd l)
					      | _ -> err ((quote_string "car") ^ " can only be applied to a list")
					   )
	| Liste(Op_car::_) -> err "wrong number of arguments"
	| Liste(Op_cdr::[l]) -> exec l env (fun _ -> function
					      | Liste(l) -> 
						  if List.tl (List.tl l) = [] then
						    cont env (List.hd (List.tl l))
						  else
						    cont env (Liste(List.tl l))
					      | _ -> err ((quote_string "cdr") ^ " can only be applied to a list")
					   )
	| Liste(Op_cdr::_) -> err "wrong number of arguments"
	| Liste(Op_cwr("a")::[lst]) -> exec (Liste(Op_car::[lst])) env cont
	| Liste(Op_cwr("d")::[lst]) -> exec (Liste(Op_cdr::[lst])) env cont
	| Liste(Op_cwr(w)::[lst]) ->
	    let string_of_char c = let s = "?" in s.[0] <- c; s in
	      exec (Liste(Op_cwr(string_of_char (w.[(String.length w)-1]))::[lst])) env
		(fun _ r -> exec (Liste(Op_cwr(String.sub w 0 ((String.length w)-1))::[r])) env cont)
	| Liste(Op_cwr(_)::_) -> err "wrong number of arguments"
	| Liste(Op_if::cond::c1::[c2]) ->
	    exec cond env (fun _ -> function
			     | Liste([]) -> exec c2 env cont
			     | _ -> exec c1 env cont
			  )
	| Liste(Op_if::_) -> err "wrong number of arguments"
	| Liste(Op_eq_int::a::[b]) ->
	    exec a env (fun _ -> function
			  | Entier(na) -> exec b env (fun _ -> function
							| Entier(nb) ->
							    if na = nb then
							      cont env (Liste([Entier 1]))
							    else
							      cont env (Liste([]))
							| _ -> err "must be applied to integers"
						     )
			  | bad -> err "must be applied to integers"
		       )
	| Liste(Op_eq_int::_) -> err "wrong number of arguments"
	| Exec_liste [] -> cont env (Liste [])
	| Exec_liste(o::s) ->
	    exec (Exec_liste s) env (fun _ rs -> exec o env (
				       fun _ r -> match rs with
					 | Liste l -> cont env (Liste (r::l))
					 | _ -> err "this error should never have happened"
				     ))
	| Liste(Op_lambda::Liste([])::[body]) -> exec body env cont
	| Liste(Op_lambda::Liste(vars)::[body]) ->
	    let rec list_of_stringListe = function
	      | [] -> []
	      | Chaine c::s -> c::(list_of_stringListe s)
		  (* Hummmmmmm *)
	      | _ -> Obj.magic (err "variables names should only be identifiers")
	    in
	    let rec all_different = function
	      | x::s -> (not (List.mem x s)) && (all_different s)
	      | [] -> true
	    in
	    let var_list = list_of_stringListe vars in
	      if not (all_different var_list) then (err "variables names should be distinct");
	      cont env (Fonction(var_list, body))
	| Liste(Op_lambda::_) -> err "wrong number of arguments"
	| Liste(Fonction([var], body)::[arg]) ->
	    exec (Liste(Op_let::Liste([Liste(Chaine var::[arg])])::[body])) env cont
	| Liste(Fonction(var::vars, body)::[arg]) ->
	    cont env (Fonction(vars, (Liste(Op_let::Liste([Liste(Chaine var::[arg])])::[body]))))
	| Liste(Fonction(var::vars, body)::arg::args) ->
	    exec (Liste(Fonction(vars, (Liste(Op_let::Liste([Liste(Chaine var::[arg])])::[body])))::args)) env cont
	| Fonction(vars, body) as f -> cont env f
	| Liste(Op_module::Chaine c::[Liste(l)]) ->
	    exec (Liste(Op_let::Liste(l)::[Liste(Op_setq::Chaine c::[Liste(Op_lambda::Liste([Chaine "x"])::[Liste(Op_eval::[Chaine "x"])])])])) env cont
	| Liste(Op_module::_) -> err "wrong number or type of arguments"
	| Liste(f::args) ->
	    exec (Exec_liste args) env (
	      fun _ e_args_liste ->
		match e_args_liste with
		  | Liste e_args -> exec f env (
		      fun env -> function
			| Fonction(_, _) as r -> exec (Liste(r::e_args)) env cont
			| Continuation(_, _) as c -> exec (Liste(c::args)) env cont (* cont will never be used *)
			| r -> err ((quote_string (string_of_obj r)) ^ " is not a function, it cannot be applied")
		    )
		  | _ -> err "this error should never have happened"
	    )
	| _ -> err "was not able to understand this"
	    
  let run code =
    let ret = ref (Entier 0) in
      exec code (Env.make ()) (fun _ r -> ret := r);
      !ret
end
